import pytest

from reactivated import forms
from sample.server.apps.samples import models


@pytest.mark.django_db
@pytest.mark.urls("tests.urls")
def test_autocomplete(client):
    composer = models.Composer.objects.create(name="Richard Wagner")
    models.Composer.objects.create(name="Wolfgang Amadeus Mozart")

    assert client.get("/autocomplete-view/").status_code == 200

    assert (
        client.post(
            "/autocomplete-view/",
            {"name": "Zarzuela", "style": "BUFFA", "composer": composer.pk},
        ).status_code
        == 302
    )

    response = client.get(
        "/autocomplete-view/", {"autocomplete": "name", "query": "Wagner"}
    )
    assert "Rendered form" in str(response.content)

    response = client.get(
        "/autocomplete-view/", {"autocomplete": "composer", "query": "Wagner"}
    )
    assert response.json()["results"][0]["label"] == "Richard Wagner"


@pytest.mark.django_db
@pytest.mark.urls("tests.urls")
def test_invalid_value(client):
    response = client.post(
        "/autocomplete-view/", {"name": "Zarzuela", "composer": "21s7"}
    )
    assert "Select a valid choice" in response.context["form"].errors["composer"][0]
    assert response.context["form"]["composer"].value() == "21s7"


@pytest.mark.django_db
@pytest.mark.urls("tests.urls")
def test_typed_autocomplete(client):
    composer = models.Composer.objects.create(name="Richard Wagner")
    models.Composer.objects.create(name="Wolfgang Amadeus Mozart")

    assert client.get("/typed-autocomplete-view/").status_code == 200

    assert (
        client.post(
            "/typed-autocomplete-view/", {"name": "Zarzuela", "composer": composer.pk}
        ).status_code
        == 302
    )

    response = client.get(
        "/typed-autocomplete-view/", {"autocomplete": "name", "query": "Wagner"}
    )
    assert "" in str(response.content)

    response = client.get(
        "/typed-autocomplete-view/", {"autocomplete": "composer", "query": "Wagner"}
    )
    assert response.json()["results"][0]["label"] == "Richard Wagner"


def test_prefix_calculation(client):
    assert forms.get_form_or_form_set_descriptor("opera_form_set-0-composer_field") == (
        "opera_form_set",
        "composer_field",
    )

    assert forms.get_form_or_form_set_descriptor("opera_form-composer_field") == (
        "opera_form",
        "composer_field",
    )

    assert forms.get_form_or_form_set_descriptor("composer_field") == (
        None,
        "composer_field",
    )



from reactivated import export, value_registry


def test_export_registry():
    FOO = 1
    BAR = 2

    export(FOO)

    export(BAR)

    assert value_registry["FOO"] == 1
    assert value_registry["BAR"] == 2



import enum
from typing import NamedTuple

import pytest
import rest_framework
from django import forms
from django.apps.registry import Apps
from django.core.exceptions import ValidationError
from django.db.models import Model

from reactivated import constraints, fields
from reactivated.forms import EnumChoiceField
from sample.server.apps.samples import models


class Member(NamedTuple):
    title: str

    def __str__(self) -> str:
        return self.title


class EnumTest(enum.Enum):
    FIRST = "First"
    SECOND = "Second"
    THIRD = "Third"
    CHUNKY = Member(title="Chunky")


def test_enum_model_form():
    class EnumForm(forms.ModelForm):
        class Meta:
            fields = ["hemisphere"]
            model = models.Continent

    form = EnumForm({"hemisphere": "NORTHERN"})
    assert form.is_valid()

    form = EnumForm(
        {"hemisphere": "NORTHERN"},
        instance=models.Continent(hemisphere=models.Continent.Hemisphere.SOUTHERN),
    )
    assert form.is_valid()
    assert form.initial["hemisphere"] == models.Continent.Hemisphere.SOUTHERN
    assert form.cleaned_data["hemisphere"] == models.Continent.Hemisphere.NORTHERN
    assert form.instance.hemisphere == models.Continent.Hemisphere.NORTHERN

    form = EnumForm({"hemisphere": "wrong"})
    assert form.is_valid() is False

    form = EnumForm(
        {"hemisphere": "wrong"},
        instance=models.Continent(hemisphere=models.Continent.Hemisphere.SOUTHERN),
    )
    assert form.is_valid() is False
    assert form.initial["hemisphere"] == models.Continent.Hemisphere.SOUTHERN

    EnumForm.base_fields["hemisphere"].disabled = True
    form = EnumForm({"hemisphere": "NORTHERN"})
    assert form.is_valid() is True

    EnumForm.base_fields["hemisphere"].disabled = True
    form = EnumForm(
        {"hemisphere": "SOUTHERN"},
        instance=models.Continent(hemisphere=models.Continent.Hemisphere.NORTHERN),
    )
    assert form.is_valid() is True
    assert form.initial["hemisphere"] == models.Continent.Hemisphere.NORTHERN
    assert form.cleaned_data["hemisphere"] == models.Continent.Hemisphere.NORTHERN
    assert form.instance.hemisphere == models.Continent.Hemisphere.NORTHERN


def test_enum_form_field():
    class EnumForm(forms.Form):
        enum_form_field = EnumChoiceField(enum=EnumTest)

    form = EnumForm({"enum_form_field": "THIRD"},)
    assert form.is_valid()
    assert form.cleaned_data["enum_form_field"] == EnumTest.THIRD

    form = EnumForm(initial={"enum_form_field": EnumTest.SECOND})
    assert form["enum_form_field"].value() == "SECOND"


def test_convert_enum_to_choices():
    (
        (first_choice, first_label),
        (second_choice, second_label),
        (_, _),
        (last_choice, last_label),
    ) = list(fields.convert_enum_to_choices(enum=EnumTest))
    assert str(first_choice) == "FIRST"
    assert first_choice.choice == EnumTest.FIRST
    assert first_label == "First"
    assert str(second_choice) == "SECOND"
    assert second_label == "Second"
    assert second_choice.choice == EnumTest.SECOND

    assert last_choice.choice == EnumTest.CHUNKY
    assert str(last_choice) == "CHUNKY"
    assert last_label == "Chunky"

    ((first_choice, first_label), *rest, (last_choice, last_label),) = list(
        fields.convert_enum_to_choices(include_blank=True, enum=EnumTest)
    )

    assert first_choice == ""
    assert first_label == "---------"

    assert last_choice.choice == EnumTest.CHUNKY
    assert str(last_choice) == "CHUNKY"
    assert last_label == "Chunky"


def test_parse_enum():
    assert fields.parse_enum(EnumTest, None) is None
    assert fields.parse_enum(EnumTest, "SECOND") is EnumTest.SECOND
    assert fields.parse_enum(EnumTest, "EnumTest.SECOND") is EnumTest.SECOND

    with pytest.raises(ValidationError, match=f"Invalid .* {EnumTest}"):
        fields.parse_enum(EnumTest, "FAKE")

    with pytest.raises(ValidationError, match=f"Invalid .* {EnumTest}"):
        fields.parse_enum(EnumTest, "EnumTest.FAKE")


def test_auto_contraint(settings):
    settings.INSTALLED_APPS = ["tests.fields"]
    test_apps = Apps(settings.INSTALLED_APPS)

    class TestModel(Model):
        enum_field = fields.EnumField(enum=EnumTest)

        class Meta:
            apps = test_apps

    constraint = TestModel._meta.constraints[0]
    assert isinstance(constraint, constraints.EnumConstraint)
    assert constraint.members == ["FIRST", "SECOND", "THIRD", "CHUNKY"]
    assert constraint.field_name == "enum_field"
    assert constraint.name == "fields_testmodel_enum_field_enum"


def test_drf_serializer(settings):
    settings.INSTALLED_APPS = ["tests.fields"]
    test_apps = Apps(settings.INSTALLED_APPS)

    class TestModel(Model):
        enum_field = fields.EnumField(enum=EnumTest)

        class Meta:
            apps = test_apps

    class TestModelSerializer(rest_framework.serializers.ModelSerializer):
        def create(self, validated_data):
            return validated_data

        def update(self, instance, validated_data):
            return validated_data

        class Meta:
            fields = "__all__"
            model = TestModel

    data = TestModelSerializer(TestModel(enum_field=EnumTest.SECOND)).data
    assert data["enum_field"] == "SECOND"

    data = TestModelSerializer({"enum_field": "SECOND"}).data
    assert data["enum_field"] == "SECOND"

    create_serializer = TestModelSerializer(data={"enum_field": "SECOND"})
    assert create_serializer.is_valid() is True
    validated_data = create_serializer.save()
    assert validated_data["enum_field"] == EnumTest.SECOND
    assert create_serializer.data["enum_field"] == "SECOND"

    update_serializer = TestModelSerializer(
        TestModel(enum_field=EnumTest.SECOND), data={"enum_field": "THIRD"}
    )
    assert update_serializer.is_valid() is True
    validated_data = update_serializer.save()
    assert validated_data["enum_field"] == EnumTest.THIRD
    assert update_serializer.data["enum_field"] == "THIRD"



import collections
from typing import Any, Sequence

from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Manager


# Mock is_simple_callable for now
# instead of the more sophisticated one from rest_framework.fields
def is_simple_callable(possible_callable: Any) -> bool:
    return callable(possible_callable)


# From Django Rest Framework 3.6
def get_attribute(instance: Any, attrs: Sequence[str]) -> Any:
    """
    Similar to Python's built in `getattr(instance, attr)`,
    but takes a list of nested attributes, instead of a single attribute.
    Also accepts either attribute lookup on objects or dictionary lookups.
    """
    for attr in attrs:
        if instance is None:
            # Break out early if we get `None` at any point in a nested lookup.
            return None
        try:
            if isinstance(instance, collections.Mapping):
                instance = instance[attr]
            else:
                instance = getattr(instance, attr)
        except ObjectDoesNotExist:
            return None

        if isinstance(instance, Manager):
            instance = instance.all()
        elif is_simple_callable(instance):
            try:
                instance = instance()
            except (AttributeError, KeyError) as exc:
                # If we raised an Attribute or KeyError here it'd get treated
                # as an omitted field in `Field.get_attribute()`. Instead we
                # raise a ValueError to ensure the exception is not masked.
                raise ValueError(
                    'Exception raised in callable attribute "{0}"; original exception was: {1}'.format(
                        attr, exc
                    )
                )

    return instance



from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Iterable,
    Literal,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

from django.core.exceptions import ValidationError
from django.db import DatabaseError, models
from django.db.models.fields import BLANK_CHOICE_DASH, NOT_PROVIDED

from .constraints import EnumConstraint

if TYPE_CHECKING:
    from django.db.models.fields import _ValidatorCallable, _ErrorMessagesToOverride
else:

    class _ValidatorCallable:
        pass

    class _ErrorMessagesToOverride:
        pass


TEnum = TypeVar("TEnum", bound=Enum)
_ST = TypeVar("_ST", bound=Enum)
_GT = TypeVar("_GT", bound=Enum)


class EnumChoice(Generic[_GT]):
    def __init__(self, choice: _GT) -> None:
        self.choice = choice

    def __str__(self) -> str:
        return self.choice.name

    def __hash__(self) -> int:
        """
        Used by DRF when creating choices in `rest_framework/fields.py`
        """
        return hash((self.choice))

    def __eq__(self, other: Union[_GT, Any, str]) -> bool:
        if isinstance(other, str):
            return str(self.choice) == other
        elif isinstance(other, Enum):
            return self.choice == other
        return False


def convert_enum_to_choices(
    *, enum: Type[Enum], include_blank: bool = False
) -> Iterable[Tuple[Union[EnumChoice[Enum], Literal[""]], str]]:
    if include_blank is True:
        yield cast(Tuple[Literal[""], str], BLANK_CHOICE_DASH[0])
    for member in enum:
        yield (EnumChoice(member), str(member.value))


class EnumChoiceIterator(Generic[_GT]):
    """ This is a special iterator that preserves the original enum. Useful so
    we can use the "choices" argument that triggers special Django behaviors,
    but leave our enum intact for reference."""

    def __init__(self, *, enum: Type[_GT], include_blank: bool = False) -> None:
        self.enum = enum
        self.include_blank = include_blank

    def __iter__(self) -> Any:
        return convert_enum_to_choices(enum=self.enum, include_blank=self.include_blank)


def coerce_to_enum(
    enum: Type[_GT], value: Union[_GT, EnumChoice[_GT], str, None]
) -> Optional[_GT]:
    if isinstance(value, EnumChoice):
        return value.choice
    elif isinstance(value, enum):
        return value
    # Narrow the type
    assert isinstance(value, str) or value is None
    return parse_enum(enum, value)


models.CharField.__class_getitem__ = classmethod(  # type: ignore[attr-defined]
    lambda cls, key: cls
)


def parse_enum(enum: Type[_GT], value: Optional[str]) -> Optional[_GT]:
    if value is None:
        return None

    for member in enum:
        # Disabled form fields will come in as Enum.Member string
        # representations. So we handle both naked member names and the full
        # representation.
        #
        # See https://code.djangoproject.com/ticket/18431 and
        # TypedChoiceField's to_python call chain.
        if member.name == value or value == str(member):
            return member

    raise ValidationError(f"Invalid input for {enum}")


class _EnumField(models.CharField[_ST, _GT]):  # , Generic[_ST, _GT]):
    # So null is handled in when adding fields through migrations.
    empty_strings_allowed = False

    def __init__(
        self,
        *,
        enum: Type[_GT],
        default: Union[Type[NOT_PROVIDED], _GT, None] = NOT_PROVIDED,
        null: bool = False,
        verbose_name: Optional[Union[str, bytes]] = None,
        unique: bool = False,
        blank: bool = False,
        db_index: bool = False,
        editable: bool = True,
        help_text: str = "",
        db_column: Optional[str] = None,
        db_tablespace: Optional[str] = None,
        validators: Iterable[_ValidatorCallable] = (),
        error_messages: Optional[_ErrorMessagesToOverride] = None,
    ):
        self.enum = enum
        self.choices = EnumChoiceIterator(enum=enum)
        self.db_collation = None

        # We skip the constructor for CharField because we do *not* want
        # MaxLengthValidator added, as our enum members do not support __len__.
        models.Field.__init__(
            self,
            choices=self.choices,
            max_length=63,
            default=default,
            null=null,
            verbose_name=verbose_name,
            unique=unique,
            blank=blank,
            db_index=db_index,
            editable=editable,
            help_text=help_text,
            db_column=db_column,
            db_tablespace=db_tablespace,
            validators=validators,
            error_messages=error_messages,
        )

    def deconstruct(self) -> Any:
        name, path, args, kwargs = super().deconstruct()
        kwargs["enum"] = self.enum
        del kwargs["max_length"]
        del kwargs["choices"]
        return name, path, args, kwargs

    def contribute_to_class(self, cls, name, **kwargs):  # type: ignore[no-untyped-def]
        """
        We don't store the enum in the constraint. Instead, we store the fields
        so the autodetection for changed enums works automatically.
        """
        super().contribute_to_class(cls, name, **kwargs)
        if "constraints" not in cls._meta.original_attrs:
            cls._meta.original_attrs["constraints"] = []

        # Note that we cannot use the constraint name interpolation syntax
        # because it's too late at this point. It's the metaclass that actually
        # interpolates the values.
        #
        # Fortunately, we can create a name dynamically.
        cls._meta.constraints.append(
            EnumConstraint(
                members=self.enum._member_names_,  # type: ignore[arg-type]
                field_name=name,
                name=f"{cls._meta.db_table}_{name}_enum",
            )
        )

    def db_type(self, connection: Any) -> str:
        if connection.settings_dict["ENGINE"] not in [
            "django.db.backends.postgresql",
            "django.db.backends.postgresql_psycopg2",
        ]:
            raise DatabaseError("EnumField is only supported on PostgreSQL")
        return super().db_type(connection)

    def from_db_value(
        self, value: Optional[str], expression: Any, connection: Any
    ) -> Optional[_GT]:
        return parse_enum(self.enum, value)

    def to_python(self, value: Union[_GT, EnumChoice[_GT], str, None]) -> Optional[_GT]:
        return coerce_to_enum(self.enum, value)

    def get_prep_value(self, value: Union[_GT, str, None]) -> Optional[str]:
        member = self.to_python(value)
        if member is None:
            return None

        return str(member.name)

    def value_to_string(self, obj: Any) -> Optional[str]:
        value = self.value_from_object(obj)
        return self.get_prep_value(value)

    def formfield(self, **kwargs: Any) -> Any:
        from .forms import EnumChoiceField

        include_blank = self.blank or not (self.has_default() or "initial" in kwargs)
        choices_with_blank = EnumChoiceIterator(
            include_blank=include_blank, enum=self.enum
        )

        # We need to pass our own choice iterator otherwise list() is called upon
        # our choices by super().formfield and we lose the enum.
        defaults = {
            **kwargs,
            "choices_form_class": EnumChoiceField,
            "choices": choices_with_blank,
        }
        return super().formfield(**defaults)


if TYPE_CHECKING:

    @overload
    def EnumField(  # type: ignore[misc]
        enum: Type[TEnum],
        default: Union[Type[NOT_PROVIDED], TEnum, None] = NOT_PROVIDED,
        null: Literal[False] = False,
        verbose_name: Optional[Union[str, bytes]] = None,
        unique: bool = False,
        blank: bool = False,
        db_index: bool = False,
        editable: bool = True,
        help_text: str = "",
        db_column: Optional[str] = None,
        db_tablespace: Optional[str] = None,
        validators: Iterable[_ValidatorCallable] = (),
        error_messages: Optional[_ErrorMessagesToOverride] = None,
    ) -> _EnumField[TEnum, TEnum]:
        ...

    @overload
    def EnumField(
        enum: Type[TEnum],
        default: Union[Type[NOT_PROVIDED], TEnum, None] = NOT_PROVIDED,
        null: Literal[True] = True,
        verbose_name: Optional[Union[str, bytes]] = None,
        unique: bool = False,
        blank: bool = False,
        db_index: bool = False,
        editable: bool = True,
        help_text: str = "",
        db_column: Optional[str] = None,
        db_tablespace: Optional[str] = None,
        validators: Iterable[_ValidatorCallable] = (),
        error_messages: Optional[_ErrorMessagesToOverride] = None,
    ) -> _EnumField[Optional[TEnum], Optional[TEnum]]:  # type: ignore[type-var]
        ...

    def EnumField(
        enum: Type[TEnum],
        default: Union[Type[NOT_PROVIDED], TEnum, None] = NOT_PROVIDED,
        null: Literal[True, False] = False,
        verbose_name: Optional[Union[str, bytes]] = None,
        unique: bool = False,
        blank: bool = False,
        db_index: bool = False,
        editable: bool = True,
        help_text: str = "",
        db_column: Optional[str] = None,
        db_tablespace: Optional[str] = None,
        validators: Iterable[_ValidatorCallable] = (),
        error_messages: Optional[_ErrorMessagesToOverride] = None,
    ) -> Union[_EnumField[TEnum, TEnum], _EnumField[Optional[TEnum], Optional[TEnum]]]:  # type: ignore[type-var]
        return _EnumField[TEnum, TEnum](enum=enum, default=default, null=null)


else:

    class EnumField(_EnumField):
        pass

    try:
        # DRF cannot serialize enums. So we do it by monkey patching. In the
        # future this may be a setting.

        # Any field with choices gets coerced by DRF to a ChoiceField,
        # regardless of its base type. So we have a special choice field to
        # look out for that.  Separately: read only fields, such as those set
        # to editable=False, will need to be mapped as well.
        from rest_framework import serializers

        class DRFEnumChoiceField(serializers.ChoiceField):
            def to_internal_value(self, data):
                data = super().to_internal_value(data)
                if isinstance(data, EnumChoice):
                    return data.choice
                return data

            def to_representation(self, obj):
                obj = super().to_representation(obj)
                if isinstance(obj, EnumChoice):
                    return obj.choice.name
                elif isinstance(obj, Enum):
                    return obj.name
                return obj

        class DRFReadOnlyEnumField(serializers.CharField):
            def to_representation(self, obj):
                return str(obj.name)

        serializers.ModelSerializer.serializer_choice_field = DRFEnumChoiceField
        serializers.ModelSerializer.serializer_field_mapping[
            EnumField
        ] = DRFReadOnlyEnumField

    except ImportError:
        pass
